/*
	File:			game.cpp

	Description:	This is the game portion of the example.  Game specific code to take the input,
					update the world, and draw is in here.  

					Code that depends on the host operating system, like PalmOS, does not belong here.  
					Instead all such functionality is obtained through the host (HostPalm.c).

					At the same time, because this code is free of the OS  it simpler to read and 
					easier to compile.

					This file is in C++ to show how to interact with HostPalm.c
					
	Author:		

	Copyright:		Copyright  2003 by Tapwave, Inc.

	Disclaimer:		IMPORTANT:  This Tapwave software is provided by Tapwave, Inc. ("Tapwave").  Your 
					use is subject to and governed by terms and conditions of the Software Development 
					Kit Agreement ("SDK Agreement") between you and Tapwave.  If you have not entered 
					into Tapwaves standard SDK Agreement with Tapwave, you have no right or license 
					to use, reproduce, modify, distribute or otherwise exploit this Tapwave software.  
					You may obtain a copy of Tapwaves standard SDK Agreement by calling 650-960-1817
					or visiting Tapwave at http://www.tapwave.com/developers/. 

	Change History (most recent first):
				
*/

/*!
This example is as simple as possible.  It shows a map that can be scrolled in
two dimensions using a joystick.  This shows

1. How to get input.
2. How to update the world.
3. How to draw.

The game can be paused and resumed as well.

*/


#include "host.h"
#include "random.h"
#include <math.h>
#include "FastGameRsc.h"



//@{
/*! \brief Map terrain types.

These terrain types all have matching bitmaps for display by the host code.
*/

const int terrain_water_shallow = 0;
const int terrain_water_deep = 1;

//@}


//@{
/*! \brief Sound types.

These sound types all have matching streams for play by the host code.
*/

const int sound_bump = 2000;

//@}


//@{
/*! \brief Rumble types.

These rumble types all have matching streams for play by the host code.
*/

const int rumble_bump = 0;

//@}

int engines_on = 0;   // tracks state of engines each frame 

/*! Everything needed for a game that scrolls a map.  The map is drawn
into an offscreen buffer, and later the visible portion of the buffer is
drawn to the screen.

All functionality to run on a machine is provided by the Host API.
*/
class Cgame
{
  private:
    //!  The width of the game map in tiles.
    int width;

    //!  The height of the game map in tiles.
    int height;

    //!  The game map.  It is allocated and initialized in the constructor.
    int *map;

    //!  The name of the game.
    char name[32];

    //!  The width of the screen, and the offscreen buffer.
    int screen_width;

    //!  The height of the screen, and the offscreen buffer.
    int screen_height;

    //@{
    /*!  The view's top left is at a fixed point offset from the top left of the screen buffer.  The 
       coordinates are 10 * tile, meaning the fraction has a range of ten and represents a portion
       of a tile. 
       So 25 is halfway in the third tile. */
    int view_x;
    int view_y;
    //@}


    //!  Paused games do not advance.  They do redraw though.
    bool paused;
	
	// DOLATER: track separate effects for both directions.  If
	// you're stuck at the top, and then move right, you should 
	// get a second rumble.
	
    // If true do some effect (sound/rumble).  Replace this 
    bool scroll_past_edge;
    
    // The effect should only be played once when the edge is reached.
    // Clear when the edge is left (by some threshold).
    bool scroll_past_edge_effect_displayed;
    
    void draw_all(void);

  public:
    Cgame(const char *gamename,
        int new_screen_width, 
        int new_screen_height,
        int random_number_seed);

    ~Cgame(void);
    
    int advance(uint32 keys, int32 joystick_x, int32 joystick_y, int32 ActionA, 
					int32 ActionB, int32 ActionC, int32 ActionD, int32 TriggerA, int32 TriggerB);
    
    void update_all(uint32 keys, int32 joystick_x, int32 joystick_y, int32 ActionA, 
					int32 ActionB, int32 ActionC, int32 ActionD, int32 TriggerA, int32 TriggerB);

    void set_pause(bool new_paused);

    bool get_pause(void);

};

Cgame *the_game = NULL;


extern void
game_set_pause(int value)
{
	if (the_game != NULL)
        the_game->set_pause(value == game_pause);
}

extern int
game_get_pause(void)
{
    int result;

	if (the_game == NULL)
		result = game_pause;
	else
		result = the_game->get_pause() ? game_pause : game_run;
	
	return result;
}


extern int
game_exists(void)
{
	return the_game != NULL;
}

extern void
game_new(int new_screen_width, int new_screen_height,
    int random_number_seed)
{
#if 0
    game_delete();
#endif
    the_game = new Cgame("", new_screen_width, new_screen_height, random_number_seed);
}


extern void
game_delete(void)
{
    if (game_exists())
    {
        delete the_game;

        the_game = NULL;
    }

}

int
game_advance(unsigned long int keys, long int joystickX, long int joystickY, 
			 long int ActionA, long int ActionB, long int ActionC, long int ActionD, 
			 long int TriggerA, long int TriggerB)
{
    return the_game->advance(keys, joystickX, joystickY, ActionA, ActionB, ActionC, 
							 ActionD, TriggerA, TriggerB);
}


/*! Create a game sized for the screen.
*/

Cgame::Cgame(const char *new_name, int new_screen_width, int new_screen_height,
    int new_random_number_seed)
{
    random_number_set_seed(new_random_number_seed);

    screen_width = new_screen_width;
    screen_height = new_screen_height;
    
    // Allocate a map for the water.
    width = 32;
    height = 32;
    map = new int[width * height];

    // Make the terrain for the water.
    for (int i = width * height - 1; i >= 0; i--)
    {
        map[i] = random_number(2) == 1 ? terrain_water_shallow : terrain_water_deep;
    }

    view_x = 0;
    view_y = 0;
    scroll_past_edge = false;
    scroll_past_edge_effect_displayed = false;
    
    paused = false;
    HostMaskKeys();

}

/*! Cleanup after the game.
*/

Cgame::~Cgame(void)
{
    delete map;
}

/*! \brief return the current pause status */

bool Cgame::get_pause(void)
{
    return paused;
}

/*! \brief Pause or unpause the game.

The effect from pausing a games is noticeable until the game advances.  A paused
game ignores the input, and does not advance.  It only redraws.  A paused game 
also may display a 
message that it is paused, or it may display a game menu.
*/


void
Cgame::set_pause(bool new_pause)
{
    paused = new_pause;
    
    if (new_pause)
        HostUnmaskKeys();
    else
        HostMaskKeys();
}


/*! \brief Draw everything in the view to the screen buffer.

The model is that the buffer is erased and then everything is drawn.
An alternative model is to undraw the objects (by restoring saved pixels) 
and then drawing all again.  The second is faster when the number of objects
is small or the background is expensive to raw.  The game could be tuned to
switch between both, depending on which should be faster.
*/

void
Cgame::draw_all(void)
{

	HostOpenScreenBuffer();

	HostDrawBackground();
// draw the stars
Draw_Stars();

// draw the powerups
Draw_Powerups();

// draw the stations
Draw_Stations();

// draw the gunships
Draw_Gunships();

// draw the rocks
Draw_Rocks();

// draw the mines
Draw_Mines();

// draw the particles
Draw_Particles();

// draw the plasma
Draw_Plasma();

// draw the player if alive
if (player_state == PLAYER_STATE_ALIVE || player_state == PLAYER_STATE_DYING)
	{	
	// set the current frame
    if (engines_on)
		wraith.curr_frame = wraith.varsI[WRAITH_INDEX_DIR] + 16*(rand()%2);
	else
		wraith.curr_frame = wraith.varsI[WRAITH_INDEX_DIR];

	// draw the bob
    NewDraw_BOB(&wraith);

    // draw the shields
    if (player_shield_count > 0)
       {
       player_shield_count--;

       // select shield frame
       wraith.curr_frame = 34-player_shield_count;    
	   
       // draw the bob
       NewDraw_BOB(&wraith);

       } // end if

	} // end if

// draw explosions last
Draw_Bursts();

// draw the score and ships left
if (huds_on)
    Draw_Info();

if (scanner_on)
    Draw_Scanner();

// last state here so it will overlay
if (player_state == PLAYER_STATE_DEAD)
   {
   // player is dead
   ready_state   = 1;
   ready_counter = 0;

   // player is done!
   Draw_Text("G A M E   O V E R",StrLen("G A M E   O V E R"),(SCREEN_WIDTH/2) - 14*(StrLen("G A M E   O V E R")>>1),SCREEN_HEIGHT/2);

   } // end if

// draw get ready?
if (ready_state==0)
   {
   // test if counter is 10 for voice
   if (ready_counter==10)
       PlayToneOS5(Entering1);

   // draw text
   Draw_Text("ENTERING  SECTOR - ALPHA 11",27,80, 180);

   // increment counter
   if (++ready_counter > 60)
      {
      // set state to ready
      ready_state = 1;
      ready_counter = 0;

      } // end if

} // end if



    if (paused)
        HostDrawGameMenu(0, 0, screen_width, screen_height);

    HostCloseScreenBuffer();
}


/*! \brief Update all game objects one time unit, including input affects.

Update the position and state of all game objects.  Handle collisions.  
Incorporate input.  

No visual changes are performed by this routine, instead that is 
handled by the draw routines.  No dirty information is kept either,
as complete redraws are done for simplicity.

*/

void
Cgame::update_all(uint32 keys, int32 joystick_x, int32 joystick_y, int32 ActionA, 
					int32 ActionB, int32 ActionC, int32 ActionD, int32 TriggerA, int32 TriggerB      //!< current joystick position
    )
{
    //int object_width;
    //int object_height;
	float xv=0;
	float yv=0;
	float plasma_xv=0;
	float plasma_yv=0;
	UInt8 temp=0;



	//put fun stuff here
	// what state is the main loop in?
if (game_state == GAME_STATE_INIT)
{
// reset everything

// transition into menu state
// game_state = GAME_STATE_MENU;

// transition into run state
game_state = GAME_STATE_RESTART;
} // end game_state is game init
else
if (game_state == GAME_STATE_MENU)
{

game_state = GAME_STATE_RESTART;

} // end if in menu state
else 
if (game_state == GAME_STATE_RESTART)
{
// start the main music
//DSound_Play(main_music_id, DSBPLAY_LOOPING);

// restart the game
Game_Reset();

// set to run state
game_state = GAME_STATE_RUNNING;


} // end if restart
else
if (game_state == GAME_STATE_RUNNING)
{




// only process player if alive
if (player_state == PLAYER_STATE_ALIVE)
{



// test if player is moving
    if (joystick_x > 7500) // right
	{
		// rotate player to right
		if (++wraith.varsI[WRAITH_INDEX_DIR] > 15)
			wraith.varsI[WRAITH_INDEX_DIR] = 0;
	
    } // end if
else
if (joystick_x < -7500) // left
    {
    // rotate player to left
    if (--wraith.varsI[WRAITH_INDEX_DIR] <  0)
	 	wraith.varsI[WRAITH_INDEX_DIR] = 15;

    } // end if

// vertical/speed motion
//AK - Up or ActionC or D
if (joystick_y < -7500 || ActionC || ActionD)  //up    
{
    // move player forward
    xv = cos_look16[wraith.varsI[WRAITH_INDEX_DIR]];
	yv = sin_look16[wraith.varsI[WRAITH_INDEX_DIR]];

    // test to turn on engines
    //if (rand()%25 == 1)
       //PlayToneOS5(Engines);

	// set engines to on
	engines_on = 1;

    Start_Particle(PARTICLE_TYPE_FADE, PARTICLE_COLOR_GREEN, 3, 
                   player_x+RAND_RANGE(-2,2), player_y+RAND_RANGE(-2,2), (-(int)(player_xv)>>3), (-(int)(player_yv)>>3));
    
    } // end if
else
if (engines_on)
   {
   // reset the engine on flag and turn off sound
   engines_on = 0;

   // turn off the sound
 
   //DSound_Stop_Sound(engines_id);
   } // end if



// toggle huds
if (TriggerA && !huds_debounce)
   {
   huds_debounce = 1;
   huds_on = (huds_on) ? 0 : 1;

   PlayToneOS5(Beep1);

   } // end if

if (!(TriggerA))
   huds_debounce = 0;

// toggle scanner
if (TriggerB && !scanner_debounce)
   {
   scanner_debounce = 1;
   scanner_on = (scanner_on) ? 0 : 1;

   PlayToneOS5(Beep1);

   } // end if

if (!(TriggerB))
   scanner_debounce = 0;



// add velocity change to player's velocity
player_xv+=xv;
player_yv+=yv;

// test for maximum velocity
vel = Fast_Distance_2D(player_xv, player_yv);

if (vel >= MAX_PLAYER_SPEED)
   {
   // re-compute velocity vector by normalizing then re-scaling
   player_xv = (MAX_PLAYER_SPEED-1)*player_xv/vel;
   player_yv = (MAX_PLAYER_SPEED-1)*player_yv/vel;
   } // end if

// add in frictional coefficient

// move player, note that these are in world coords
player_x+=player_xv;
player_y+=player_yv;

// wrap player to universe
if (player_x > UNIVERSE_MAX_X)
   {
   // reset pos
   player_x = UNIVERSE_MIN_X;

   // set warp flag
   } // end if
else
if (player_x < UNIVERSE_MIN_X)
   {
   // reset pos
   player_x = UNIVERSE_MAX_X;

   // set warp flag
   } // end if

// y coords
if (player_y > UNIVERSE_MAX_Y)
   {
   // reset pos
   player_y = UNIVERSE_MIN_Y;

   // set warp flag
   } // end if
else
if (player_y < UNIVERSE_MIN_Y)
   {
   // reset pos
   player_y = UNIVERSE_MAX_Y;

   // set warp flag
   } // end if


// test if player is firing

if (ActionA || ActionB)
	{
    if (player_ammo > 0)
       {
    	// compute plasma velocity vector
        plasma_xv = cos_look16[wraith.varsI[WRAITH_INDEX_DIR]]*PLASMA_SPEED;
	    plasma_yv = sin_look16[wraith.varsI[WRAITH_INDEX_DIR]]*PLASMA_SPEED;
 
	    // fire the plasma taking into consideration ships's velocity
        Fire_Plasma(player_x,player_y,
		            player_xv+plasma_xv,
				    player_yv+plasma_yv, PLASMA_ANIM_PLAYER);

        // use up some ammo
        if (player_ammo > 0)
        player_ammo--;

	    } // end if
    else
        {
        PlayToneOS5(Empty);
        } // end else

    } // end if

// do bounds check


// regenerate player
if (++player_regen_counter >= PLAYER_REGEN_COUNT)
    {
    // regenerate the player a bit
    if (player_damage >0 )
        player_damage--;

    // reset counter
    player_regen_counter = 0;

    } // end if

// regenerate player
if (++ammo_regen_counter >= AMMO_REGEN_COUNT)
    {
    // regenerate the player a bit
    if (player_ammo <1000 )
        player_ammo++;

    // reset counter
    ammo_regen_counter = 0;

    } // end if


// test for dying state transition
if (player_damage >= 100 && player_state == PLAYER_STATE_ALIVE)
    {
    // set damage to 100
    player_damage = 100;

    // kill player
    player_state = PLAYER_STATE_DYING;
    player_ships--;

    // set counter to 0
    player_counter = 0;

    // turn engines off
	engines_on = 0;

    // start scream
    PlayToneOS5(Breakup);
    
    } // end if


} // end if player alive
else
if (player_state == PLAYER_STATE_DYING)
   { 
   // player is dying

   // start random bursts
   int bwidth  = 16 + rand()%64;
   
   if ((rand()%4)==1)
        Start_Burst(player_x+rand()%40, player_y+rand()%8,
                   bwidth, (bwidth >> 2) + (bwidth >> 1),
                   -4+rand()%8,2+rand()%4);    

   // get jiggie with it
   wraith.varsI[WRAITH_INDEX_DIR] += (RAND_RANGE(-1,1));

   if (wraith.varsI[WRAITH_INDEX_DIR] > 15)
       wraith.varsI[WRAITH_INDEX_DIR] = 0;
   else
   if (wraith.varsI[WRAITH_INDEX_DIR] < 0)
      wraith.varsI[WRAITH_INDEX_DIR] = 15;

   // update state counter
   if (++player_counter > 150)
      {
      // set state to invincible  
      player_state = PLAYER_STATE_INVINCIBLE;

      // reset counter
      player_counter = 0;

      // reset damage
      player_damage = 0;
      } // end if


   } // end if
else
if (player_state == PLAYER_STATE_INVINCIBLE)
   {
   // player is invincible
 
   // flicker shields and play energizing sound

   // update state counter
   if (++player_counter > 70)
      {
      // set state to invincible  
      player_state = PLAYER_STATE_ALIVE;

      // reset counter
      player_counter = 0;

      // reset player position
      player_x = 0;
      player_y = 0;
      player_dx=0;
      player_dy=0; 
      player_xv=0; 
      player_yv=0;
      wraith.varsI[WRAITH_INDEX_DIR] = 0;

      // decrease player ships
      if (player_ships == 0)
         {
         // change state to dead, reset all vars
         player_state   = PLAYER_STATE_DEAD;
         player_counter = 0;
         ready_state    = 1;
         ready_counter  = 0;

         // turn engines
         //DSound_Stop_Sound(engines_id);

         // play game over sound
         PlayToneOS5(GameOver);

         } // end if       

      } // end if

    // start ready again
    if ((player_counter == 20) && (player_ships > 0))
      {
      // reset to ready
      ready_counter = 0, 
      ready_state   = 0;
      } // end if
   
   } // end if
}
if (player_state == PLAYER_STATE_DEAD)
{
	if (ActionA || ActionB || ActionC || ActionD)
		game_state = GAME_STATE_RESTART;

}

// end player sub states

// move the powerups 
Move_Powerups();

// move the stations
Move_Stations();

// move the gunships
Move_Gunships();

// move the mines
Move_Mines();

// move the asteroids
Move_Rocks();

// move the stars
Move_Stars();

// move the particles
Process_Particles();

// move the explosions
Move_Bursts();

// move the plasma
Move_Plasma();

if (!CheckMusicPlaying())
	startMusic(true);

 

    // The idea here to allow the edge notification effect once per edge visit.
    // The threshold for when an edge is no longer visited is currently one tile of distance.
	/*if (Collision_BOBS(ship,black_hole))
	{
		if (!scroll_past_edge_effect_displayed)
			scroll_past_edge = true;
	}
	else
		scroll_past_edge_effect_displayed = false;	
	*/


}


/*! \brief Advance the game one cycle

This performs the normal work to advance the game.  It takes the input
and updates the world.  It then draws the results in an offscreen buffer.
The offscreen buffer is copied to the screen at the start of the next cycle.
By drawing at a consistent time interval the results appear smooth.  Some 
input latency is added because
the results appear one time period after the input is taken, but it is not 
noticeable.

*/

int
Cgame::advance(uint32 keys,  //!< mask of keys currently pressed for this cycle
    int32 joystick_x,       //!< current joystick position
    int32 joystick_y,        //!< current joystick position
    int32 ActionA,
    int32 ActionB,
    int32 ActionC,
    int32 ActionD,
    int32 TriggerA,
    int32 TriggerB
	)
{
    /*int object_width;
    int object_height;
    int view_offset_x;
    int view_offset_y;
    */
    
    // get input
    
  
	/*

    // draw buffer
    object_height = HostGetBitmapHeight(terrain_water_shallow);
    object_width = HostGetBitmapWidth(terrain_water_shallow);
    view_offset_x = (view_x % 10 * object_width) / 10;
    view_offset_y = (view_y % 10 * object_height) / 10;
	*/
    HostDrawBuffer(0, 0, 0, 0, screen_width, screen_height);

	 // Send rumble effects.
    if (scroll_past_edge)
    {
        if (!scroll_past_edge_effect_displayed)
        	HostPlayRumble(rumble_bump);

        scroll_past_edge = false;
        scroll_past_edge_effect_displayed = true;
    }
    

    // update world.
    if (!paused)
    {
        update_all(keys, joystick_x, joystick_y, ActionA, ActionB, ActionC, ActionD, TriggerA, TriggerB);
    }


    // draw world.
    draw_all();

    // true get called again.
    return true;
}
